
// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "inverse.h"

// includes, project
#include </root/NVIDIA_CUDA_SDK/common/inc/cutil.h>

// includes, kernels
#include <inverse_kernel.cu>
////////////////////////////////////////////////////////////////////////////////
// declaration, forward
void runTest(float**,float**,int);
void randomInit(double*, int);
void printDiff(float*, float*, int, int);
void inverse(float**,float**,int);
void inverseD(double**,double**,int);


////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int
main(int argc, char** argv) {
    
CUT_DEVICE_INIT();
    clock_t start_time;
    int num_gates = 10000;
    int split_amount = 1000;
    int pass_count = 0;
    int i,j,k;
    int size = BLOCK_SIZE * BLOCK_SIZE;
    int mem_size = sizeof(float) * size;
    int mem_sizeD = sizeof(double) * size;
    double ** h_inputD = (double**) malloc(num_gates*
                                           sizeof(double*));
    double *** input_reorderedD = (double***)
                        malloc(num_gates*sizeof(double**));
    float ** h_input;
    CUDA_SAFE_CALL(cudaMallocHost((void**)&h_input,num_gates*
                                           sizeof(float*)));
    float *** input_reordered = (float***)
                        malloc(num_gates*sizeof(float**));
    float ** results_single = (float**) malloc
                              (num_gates*sizeof(float*)); 
    float ** results_temp = (float **) malloc(size*
                                          sizeof(float*));
    double ** resultsD_temp = (double**) malloc(size*
                                          sizeof(double*));
    double ** results_double = (double**) malloc
                              (num_gates*sizeof(double*));
    float ** results_div = (float**) malloc
                              (num_gates*sizeof(float*));
    float ** div_in_array;
    float ** h_in_array;
    CUDA_SAFE_CALL(cudaMallocHost((void**)&h_in_array,
                              (split_amount*sizeof(float*))));
    
    srand(2008);
   
    CUDA_SAFE_CALL(cudaMalloc((void**) &div_in_array,
                              sizeof(float*)*split_amount));

    for (i = 0; i < num_gates; i++) {
       CUDA_SAFE_CALL(cudaMallocHost((void**)&h_input[i],mem_size));
       input_reordered[i] = (float**) malloc
                               (BLOCK_SIZE*sizeof(float*));
       h_inputD[i] = (double*) malloc(mem_sizeD);
       input_reorderedD[i] = (double**) malloc
                               (BLOCK_SIZE*sizeof(double*));
       results_single[i] = (float*) malloc(mem_size);
       results_double[i] = (double*) malloc(mem_sizeD);
       results_div[i] = (float*) malloc(mem_size);
       randomInit(h_inputD[i], size);
       if (i < split_amount) {
          CUDA_SAFE_CALL(cudaMalloc((void**) &h_in_array[i],mem_size));
       }
    }

    CUDA_SAFE_CALL(cudaMemcpy(div_in_array, h_in_array, split_amount * sizeof(float*), cudaMemcpyHostToDevice));

    for (i = 0; i < BLOCK_SIZE; i++) {
       results_temp[i] = (float *) malloc(size*
                                            sizeof(float));
       resultsD_temp[i] = (double *) malloc(size*
                                            sizeof(double));
    }

    for (i = 0;i < num_gates; i++) {
       for (j = 0;j < BLOCK_SIZE; j++) {
          input_reordered[i][j] = (float *) malloc(BLOCK_SIZE
                                           *sizeof(float));
          input_reorderedD[i][j] = (double *) malloc
                               (BLOCK_SIZE*sizeof(double));
       }
    }

    for (i = 0; i < num_gates; i++) {
       for (j = 0;j < BLOCK_SIZE; j++) {
          for (k = 0;k < BLOCK_SIZE; k++) {
            h_input[i][j*BLOCK_SIZE+k] = float(h_inputD[i]
                                          [j*BLOCK_SIZE+k]);
            input_reordered[i][k][j] = float (h_inputD[i]
                                          [j*BLOCK_SIZE+k]);
            input_reorderedD[i][k][j] = h_inputD[i]
                                          [j*BLOCK_SIZE+k];
          }
       }
    }
    
    start_time = clock();
    for (i = 0; i < num_gates; i++) {
       inverse(input_reordered[i],results_temp,
                                             BLOCK_SIZE);
       for (j=0;j<BLOCK_SIZE;j++) {
          for (k=0;k<BLOCK_SIZE;k++) {
             results_single[i][j*BLOCK_SIZE+k] = results_temp[k][j];
          }
       }
    }
    printf("Time Single: %f seconds\n",(double)(clock()-start_time)/CLOCKS_PER_SEC);

    start_time = clock();
    for (i = 0; i < num_gates; i++) {
       inverseD(input_reorderedD[i],resultsD_temp,
                                             BLOCK_SIZE);
       for (j=0;j<BLOCK_SIZE;j++) {
          for (k=0;k<BLOCK_SIZE;k++) {
             results_double[i][j*BLOCK_SIZE+k] = results_temp[k][j];
          }
       }
    }
    
    printf("Time Double: %f seconds\n",(double)(clock()-start_time)/CLOCKS_PER_SEC);

    start_time = clock();
    for (int j = 0; j < num_gates/split_amount; j++) {
       for (int i = 0; i < num_gates; i++) {
          CUDA_SAFE_CALL(cudaMemcpy(h_in_array[i], h_input[i + 
                 j*split_amount], mem_size, cudaMemcpyHostToDevice));
       }
       
       runTest(div_in_array,h_in_array,split_amount);
       for (int i = 0; i < num_gates; i++) {
          CUDA_SAFE_CALL(cudaMemcpy(results_div[i+ j*split_amount],h_in_array[i],BLOCK_SIZE*BLOCK_SIZE*sizeof(float), cudaMemcpyDeviceToHost));
       }
    }
    //printf("Time CUDA: %f seconds\n",(double)(clock()-start_time)/CLOCKS_PER_SEC);   

    for (i=0; i<num_gates;i++) {
      CUTBoolean res = cutCompareL2fe(results_single[i], results_div[i], size, 1e-5f);
    if (i == 0)
    printf("Time CUDA: %f seconds\n",(double)(clock()-start_time)/CLOCKS_PER_SEC); 
    //printf("Test %s \n", (1 == res) ? "PASSED" : "FAILED");
    if (res!=1) printDiff(results_single[i], results_div[i], BLOCK_SIZE, BLOCK_SIZE); 
    else
      pass_count++;
    }
    printf("\nPasses: %d / %d",pass_count,num_gates);
    printf("\n");
 /*for (i=0;i<BLOCK_SIZE;i++) {
       for (j=0;j<BLOCK_SIZE;j++) {
          printf("%8.8lf ",h_input[0][i*BLOCK_SIZE+j]);
       }
       printf("\n");
    }
printf("\n");   */
    for (i=0; i < BLOCK_SIZE; i++) {
       for (j =0;j<BLOCK_SIZE;j++) {
             free(input_reordered[i][j]);
       }
       free(results_temp[i]);
    }
    for (i =0;i<num_gates;i++) {
       free(results_double[i]);   
       CUDA_SAFE_CALL(cudaFreeHost(h_input[i]));
       free(results_single[i]);
       free(input_reordered[i]);
       if (i < split_amount) {
          CUDA_SAFE_CALL(cudaFree(h_in_array[i]));
       } 
   }
   CUDA_SAFE_CALL(cudaFreeHost(h_in_array));
   free(results_temp);
   free(results_double);   
   CUDA_SAFE_CALL(cudaFreeHost(h_input));
   free(results_single);
   free(input_reordered);
   CUDA_SAFE_CALL(cudaFree(div_in_array));
   CUT_EXIT(argc, argv);
}
////////////////////////////////////////////////////////////////////////////////
//! Run a simple test for CUDA
////////////////////////////////////////////////////////////////////////////////
void
runTest(float ** div_in_array,float** h_in_array,int num_arrays)
{
   
    // setup execution parameters
    dim3 threads(8,10);
    dim3 grid(num_arrays/10);
    
    // execute the kernel
    inverse<<< grid, threads >>>(div_in_array,BLOCK_SIZE,num_arrays);
    
    // check if kernel execution generated and error
    CUT_CHECK_ERROR("Kernel execution failed");



}

// Allocates a matrix with random float entries.
void randomInit(double* data, int size)
{
    for (int i = 0; i < size; ++i)
        data[i] = (double)rand() / (double)RAND_MAX;
}

void printDiff(float *data1, float *data2, int width, int height)
{
  int i,j,k;
  int error_count=0;
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      k = j*width+i;
      if (data1[k] != data2[k]) {
         printf("diff(%d,%d) CPU=%4.8f, GPU=%4.8f n", i,j, data1[k], data2[k]);
         error_count++;
      }
    }
  }
  printf(" nTotal Errors = %d n", error_count);
}
